@@ -1,9 +1,9 @@
-# $Id: XPointer.pm,v 1.11 2004/11/15 05:22:13 asc Exp $
+# $Id: XPointer.pm,v 1.13 2004/11/16 03:56:18 asc Exp $
 use strict;
 
 package Apache::XPointer;
 
-$Apache::XPointer::VERSION = '1.01';
+$Apache::XPointer::VERSION = '1.1';
 
 =head1 NAME
 
@@ -16,6 +16,8 @@ Apache::XPointer - mod_perl handler to address XML fragments.
   <FilesMatch "\.xml$">
    SetHandler	perl-script
    PerlHandler	Apache::XPointer::XPath
+
+   PerlSetVar   XPointerSendRangeAs    "application/xml"
   </FilesMatch>
 
  </Directory>
@@ -24,15 +26,17 @@ Apache::XPointer - mod_perl handler to address XML fragments.
 
  my $ua  = LWP::UserAgent->new();
  my $req = HTTP::Request->new(GET => "http://example.com/foo/bar/baz.xml");
- $req->header("Range" => qq(xmlns("x=x-urn:example")xpointer(*//x:thingy)));
+
+ $req->header("Range"  => qq(xmlns("x=http://example.com#")xpointer(*//x:thingy)));
+ $req->header("Accept" => qq(application/xml));
 
  my $res = $ua->request($req);
 
 =head1 DESCRIPTION
 
 Apache::XPointer is a mod_perl handler to address XML fragments using
-the HTTP 1.1 I<Range> header, as described in the paper : I<A Semantic
-Web Resource Protocol: XPointer and HTTP>.
+the HTTP 1.1 I<Range> and I<Accept> headers and the XPath scheme, as described
+in the paper : I<A Semantic Web Resource Protocol: XPointer and HTTP>.
 
 Additionally, the handler may also be configured to recognize a conventional
 CGI parameter as a valid range identifier.
@@ -40,6 +44,10 @@ CGI parameter as a valid range identifier.
 If no 'range' property is found, then the original document is
 sent unaltered.
 
+If an I<Accept> header is specified with no corresponding match, then the
+server will return (406) HTTP_NOT_ACCEPTABLE. Successful queries will return
+(206) HTTP_PARTIAL_CONTENT.
+
 =head1 IMPORTANT
 
 This package is a base class and not expected to be invoked
@@ -57,9 +65,7 @@ Consult L<Apache::XPointer::RDQL>
 
 =head1 MOD_PERL COMPATIBILITY
 
-This handler will work with both mod_perl 1.x and mod_perl 2.x; it
-works better in 1.x because it supports Apache::Request which does
-a better job of parsing CGI parameters.
+This handler will work with both mod_perl 1.x and mod_perl 2.x.
 
 =cut
 
@@ -80,7 +86,11 @@ BEGIN {
          require Apache::URI;
          require APR::Table;
          require APR::URI;
-         Apache::Const->import(-compile => qw(OK DECLINED HTTP_NOT_FOUND HTTP_INTERNAL_SERVER_ERROR));
+         require CGI;
+
+         Apache::Const->import(-compile => qw(OK DECLINED HTTP_NOT_FOUND HTTP_NOT_ACCEPTABLE HTTP_PARTIAL_CONTENT HTTP_INTERNAL_SERVER_ERROR));
+
+         CGI->compile(qw(param));
       }
 
      else {
@@ -88,7 +98,12 @@ BEGIN {
          require Apache::Constants;
          require Apache::Log;
          require Apache::Request;
-         Apache::Constants->import(qw(OK DECLINED NOT_FOUND SERVER_ERROR));
+
+         # mod_perl 1.x does not declare a 
+         # HTTP_PARTIAL_CONTENT constant so
+         # it gets hard-coded below
+
+         Apache::Constants->import(qw(OK DECLINED NOT_FOUND HTTP_NOT_ACCEPTABLE SERVER_ERROR));
      }
 }
 
@@ -96,36 +111,43 @@ sub handler : method {
   my $pkg    = shift;
   my $apache = shift;
 
-  my $range = $pkg->_header_in($apache,"Range");
-
-  if ((! $range) && ($apache->dir_config("XPointerAllowCGIRange") =~ /^on$/i)) {
-      my $rparam  = $apache->dir_config("XPointerCGIRangeParam") || "range";
-
-      # Waiting for Apache::Request to be ported
-      # to mod_perl2 because default query parser
-      # doesn't do a very good job of handling stuff
-      # like range=xmlns(foo=x-urn:bar)
-
-      if ($pkg->_mp2()) {
-	  $apache->parsed_uri()->query() =~ /^$rparam=(.*)$/;
-	  $range = $1;
-      }
-
-      else {
-	  my $request = Apache::Request->new($apache);
-	  $range = $request->param($rparam);
-      }
-  }
+  my $range  = $pkg->range($apache);
 
   if (! $range) {
       return $pkg->_declined();
   }
 
+  my $accept = undef;
+
+  if (my $possible = $pkg->accept($apache)) {
+
+      foreach my $choice (split(",",$possible)) {
+	  
+	  $choice =~ s/^\s+//;
+	  $choice =~ s/\s+$//;
+
+          # hand-waving...levels
+	  # ...more hand-waving
+
+	  $choice =~ s/;.*//;
+
+	  if ($pkg->send_as($choice)) {
+	      $accept = $choice;
+	      last;
+	  }
+      }
+
+      if (! $accept) {
+	  $apache->log()->error("unable to send request as '$accept'");
+	  return $pkg->_not_acceptable();
+      }
+  }
+
   #
 
-  my ($ns,$pointer) = $pkg->parse_range($apache,$range);
+  my $parsed = $pkg->parse_range($apache,$range);
 
-  if (! $pointer) {
+  if (! $parsed) {
       $apache->log()->error(sprintf("failed to parse range '%s'",
 				    $range));
       
@@ -134,14 +156,106 @@ sub handler : method {
 
   #
 
-  my $res = $pkg->range($apache,$ns,$pointer);
+  my $res = $pkg->query($apache,$parsed);
 
   if ((! $res) || (! $res->{success})) {
       return $res->{response};
   }
 
-  $pkg->send_results($apache,$res);
-  return $pkg->_ok();
+  #
+
+  my $ok = $pkg->send_results($apache,$res,$accept);
+  return $ok;
+}
+
+sub send_results {
+    my $pkg    = shift;
+    my $apache = shift;
+    my $res    = shift;
+    my $accept = shift;
+
+    $accept ||= $apache->dir_config("XPointerSendRangeAs");
+    
+    my $method = $pkg->send_as($accept);
+
+    if (! $pkg->can($method)) {
+	$apache->log()->error("Unknown send as method '$method'");
+	return $pkg->_server_error();
+    }
+
+    #
+
+    $pkg->send_headers($apache,$res,$accept);
+    $pkg->$method($apache,$res);
+
+    # We set the status in 'send_headers' because
+    # apache/mod_perl keeps trying to send back an
+    # HTML error page whenever we return 206...
+
+    return 0;
+}
+
+sub send_headers {
+    my $pkg    = shift;
+    my $apache = shift;
+    my $res    = shift;
+    my $accept = shift;
+
+    $apache->status($pkg->_partial_content());
+
+    $pkg->_header_out($apache,"Content-Base",
+		      sprintf("%s:%s",$apache->hostname(),$apache->get_server_port()));
+
+    $pkg->_header_out($apache,"Content-Location",
+		      $apache->uri());
+
+    $pkg->_header_out($apache,"Content-Range",
+		      $pkg->range($apache));
+
+    $apache->content_type($accept);
+
+    if (! $pkg->_mp2()) {
+	$apache->send_http_header();
+    }
+
+    return 1;
+}
+
+sub range {
+    my $pkg    = shift;
+    my $apache = shift;
+
+    my $range = $pkg->_header_in($apache,"Range");
+    
+    if ((! $range) &&
+	($apache->dir_config("XPointerAllowCGI") =~ /^on$/i)) {
+
+	my $rparam  = $apache->dir_config("XPointerCGIRangeParam") || "range";
+	$range      = $pkg->_param($apache,$rparam);
+    }
+    
+    return $range;
+}
+
+sub accept {
+    my $pkg = shift;
+    my $apache = shift;
+
+    my $accept = $pkg->_header_in($apache,"Accept");
+
+    if ((! $accept) &&
+	($apache->dir_config("XPointerAllowCGI") =~ /^on$/i)) {
+
+	my $aparam  = $apache->dir_config("XPointerCGIAcceptParam") || "accept";		
+	$accept     = $pkg->_param($apache,$aparam);
+    }
+
+    return $accept;
+}
+
+sub send_as {
+    my $pkg = shift;
+    return $pkg->_nometh(@_);
 }
 
 sub parse_range {
@@ -149,20 +263,28 @@ sub parse_range {
     return $pkg->_nometh(@_);
 }
 
-sub range {
+sub query {
     my $pkg = shift;
     return $pkg->_nometh(@_);
 }
 
-sub send_results {
-    my $pkg    = shift;
-    return $pkg->_nometh(@_);
-}
-
 sub _mp2 {
     return MP2;
 }
 
+sub _param {
+    my $pkg    = shift;
+    my $apache = shift;
+    my $field  = shift;
+
+    if ($pkg->_mp2()) {
+	return CGI::param($field);
+    }
+    
+    my $request = Apache::Request->new($apache);
+    return $request->param($field);
+}
+
 sub _nometh {
     my $pkg    = shift;
     my $apache = shift;
@@ -190,6 +312,16 @@ sub _not_found {
     return ($pkg->_mp2()) ? Apache::HTTP_NOT_FOUND() : Apache::Constants::NOT_FOUND();
 }
 
+sub _not_acceptable {
+    my $pkg = shift;
+    return ($pkg->_mp2()) ? Apache::HTTP_NOT_ACCEPTABLE() : Apache::Constants::HTTP_NOT_ACCEPTABLE();
+}
+
+sub _partial_content {
+    my $pkg = shift;
+    return ($pkg->_mp2()) ? Apache::HTTP_PARTIAL_CONTENT() : 206;
+}
+
 sub _ok {
     my $pkg = shift;
     return ($pkg->_mp2()) ? Apache::OK() : Apache::Constants::OK();
@@ -214,11 +346,11 @@ sub _header_out {
 
 =head1 VERSION
 
-1.01
+1.1
 
 =head1 DATE
 
-$Date: 2004/11/15 05:22:13 $
+$Date: 2004/11/16 03:56:18 $
 
 =head1 AUTHOR
 
